/*
 * tclUnixThrd.h --
 *
 *      This header file defines things for thread support.
 *
 * Copyright (c) 1998 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#)
 */
 
#ifndef _TCLMACTHRD
#define _TCLMACTHRD

#ifdef TCL_THREADS

#endif /* TCL_THREADS */
#endif /* _TCLMACTHRD */
